# -*- coding: utf-8 -*-

class TreeError(Exception):
    def __init__(self,value):
        self.value = 'Tree error: '+value
    def __str__(self):
        return repr(self.value)

class InvalidPattern(Exception):
    def __init__(self,value):
        self.value = 'Invalid pattern: '+value
    def __str__(self):
        return repr(self.value)



# -*- coding: utf-8 -*-

import pgf
from PIL import Image
import hashlib
import graphviz
import os
# from screeninfo import get_monitors

EMPTY="?"
SYMB="symb"

def get_image(dot_code,tree_str,suffix='',filename=None,dir='.',format = 'png'):
    if not filename:
        hash_object = hashlib.md5(tree_str.encode())
        filename_base = hash_object.hexdigest()+suffix
    else:
        filename_base = filename
    dot_filename = os.path.join(dir,filename_base + '.dot')
    img_filename = dot_filename + '.' + format
    dot_graph = graphviz.Source(dot_code)
    dot_graph.format = format
    dot_graph.render(dot_filename)
    return img_filename

def _show_img(dot_code,tree_str,suffix='',filename='',dir='.'):
    png_filename = get_image(dot_code,tree_str,suffix,filename,dir,'png')
    print(png_filename)
    img = Image.open(png_filename)

    # width,height = img.size
    # m = get_monitors()
    # m_height = m[0].height
    # factor = (m_height*.8) / height
    #
    # img = img.resize((int(width*factor), int(height*factor)))
    img.show()

def _show_abs_image(tree,grammar,filename='',dir='.'):
    dot_code = grammar.graphvizAbstractTree(tree)
    tree_str = str(tree)
    _show_img(dot_code,tree_str,'_abs',filename,dir)

def _show_parse_image(tree,concrete,filename='',dir='.'):
    dot_code = concrete.graphvizParseTree(tree)
    tree_str = str(tree)
    _show_img(dot_code,tree_str,'_parse',filename,dir)

def get_abs_image(tree,grammar,filename=None,dir='.',format = 'png'):
    dot_code = grammar.graphvizAbstractTree(tree)
    tree_str = str(tree)
    return get_image(dot_code,tree_str,'_abs',filename,dir,format)

def get_parse_image(tree,concrete,filename=None,dir='.',format = 'png'):
    dot_code = concrete.graphvizParseTree(tree)
    tree_str = str(tree)
    return get_image(dot_code,tree_str,'_parse',filename,dir,format)

def node_is_empty(tree):
    tree = pgf.readExpr(str(tree))
    if not tree.unpack()[0]:
        return True
    return False

def is_equal(tree1,tree2):
    if str(tree1) == str(tree2):
        return True
    return False

def depth(tree):
    tree = pgf.readExpr(str(tree))
    children = children_trees(tree)
    if len(children) == 0:
        return 0
    children_lens = []
    for c in children:
        children_lens.append(depth(c))
    return max(children_lens) + 1

def root_str(tree):
    tree = pgf.readExpr(str(tree))
    try:
        (fun_str,children) = tree.unpack()
        return fun_str
    except ValueError:
        return str(tree)
    except AttributeError:
        return str(tree)
    except TypeError:
        return str(tree)

def root_cat(tree,grammar):
    tree = pgf.readExpr(str(tree))
    try:
        fun_str = root_str(tree)
        if fun_str:
            fun_type = grammar.functionType(fun_str)
            return fun_type.cat
        else:
            return EMPTY
    except KeyError:
        return SYMB

def children_trees(tree):
    tree = pgf.readExpr(str(tree))
    try:
        (fun_str,children) = tree.unpack()
        return children
    except ValueError:
        return []
    except AttributeError:
        return []
    except TypeError:
        return []

def subtrees_of_cat(tree,cat,grammar,overlap=True):
    tree = pgf.readExpr(str(tree))
    trees = []
    children = children_trees(tree)
    if root_cat(tree,grammar) == cat:
        if overlap:
            trees.append(tree)
        else:
            return [tree]
    if len(children) == 0:
        return trees
    for c in children_trees(tree):
        trees += subtrees_of_cat(c,cat,grammar,overlap)
    return trees

def subtrees_of_fun(tree,fun,overlap=True):
    tree = pgf.readExpr(str(tree))
    trees = []
    children = children_trees(tree)
    if root_str(tree) == fun:
        if overlap:
            trees.append(tree)
        else:
            return [tree]
    if len(children) == 0:
        return trees
    for c in children_trees(tree):
        trees += subtrees_of_fun(c,fun,overlap)
    return trees

def sanity_check_tree(tree,grammar):
    tree = pgf.readExpr(str(tree))
    try:
        f_cat = root_cat(tree,grammar)
        grammar.checkExpr(tree,pgf.readType(f_cat))
        return True
    except Exception as e:
        return False

def leaf_function_names_by_cat(cat,grammar):
    function_names = grammar.functionsByCat(cat)
    leaf_function_names = []
    for f in function_names:
        fun_type = grammar.functionType(f)
        args,cat,x = fun_type.unpack()
        if len(args) == 0:
            leaf_function_names.append(f)
    return leaf_function_names

def alternative_leaf_function_names(tree,grammar,conc_name):
    concrete = grammar.languages[conc_name]
    tree = pgf.readExpr(str(tree))
    fun_name = root_str(tree)
    fun_cat = root_cat(tree,grammar)
    leaf_nodes = leaf_nodes_with_ids(tree)
    all_funs = leaf_function_names_by_cat(fun_cat,grammar)
    all_leaf_funs = [f for f in all_funs if depth(f) == 0 and concrete.hasLinearization(f)]
    other_leaf_funs = [f for f in all_leaf_funs if f != fun_name]
    return [fun_name] + other_leaf_funs

def _leaf_nodes_with_ids(tree,prev_id=-1):
    tree = pgf.readExpr(str(tree))
    children = children_trees(tree)
    leaves = []
    if len(children) == 0: # tree is a leaf
        cur_id = prev_id + 1
        return [(tree,cur_id)],cur_id
    for c in children:
        new_leaves,prev_id = _leaf_nodes_with_ids(c,prev_id)
        leaves += new_leaves
        cur_id = prev_id + 1
    return leaves,cur_id

def leaf_nodes_with_ids(tree):
    leaves,root_id = _leaf_nodes_with_ids(tree,-1)
    return leaves

def _empty_node_ids(tree,prev_id=-1):
    tree = pgf.readExpr(str(tree))
    empty_ids = []
    if str(tree) == EMPTY:
        cur_id = prev_id + 1
        return empty_ids + [cur_id],cur_id
    children = children_trees(tree)
    if len(children) == 0: # tree is a leaf
        cur_id = prev_id + 1
        return empty_ids,cur_id
    for c in children:
        new_ids,prev_id = _empty_node_ids(c,prev_id)
        empty_ids += new_ids
        cur_id = prev_id + 1
    return empty_ids,cur_id

def empty_node_ids(tree):
    empty_ids,root_id = _empty_node_ids(tree,-1)
    return empty_ids

def _path_to_root(tree,id,prev_id=-1,path=[]):
    tree = pgf.readExpr(str(tree))
    children = children_trees(tree)
    if len(children) == 0: # leaf node
        cur_id = prev_id + 1
        if cur_id == id:
            path.append((root_str(tree),cur_id))
            # print(f'path is: {path}')
    else:
        parent_state = [p for p in path]
        for c in children:
            path,prev_id = _path_to_root(c,id,prev_id,path)
            cur_id = prev_id + 1
        if tuple(parent_state) != tuple(path):
            path.append((root_str(tree),cur_id))
        if cur_id == id:
            path.append((root_str(tree),cur_id))

    return path,cur_id

def path_to_root(tree,id):
    path,root_id = _path_to_root(tree,id,prev_id=-1,path=[])
    return path

def size(tree,prev_id=-1):
    tree = pgf.readExpr(str(tree))
    children = children_trees(tree)
    if len(children) == 0: # leaf node
        cur_id = prev_id + 1
    else:
        for c in children:
            prev_id = size(c,prev_id)
            cur_id = prev_id + 1

    return cur_id



# -*- coding: utf-8 -*-

import pytest
import pgf

@pytest.fixture
def multilexlang_2022_07():
    return {
        "grammar": "tests/MultiLexLang.pgf",
        "conc_name": "MultiLexLangZul",
        "exprs": [
            pgf.readExpr('PhrUtt NoPConj (UttS (UseCl TPresTemp PPos (PredVP (DetCN (DetNum NumSg) (UseN boy_N)) (UseV walk_V)))) NoVoc'),
            pgf.readExpr('PhrUtt NoPConj (UttS (UseCl TFutTemp PPos (PredVP (DetCN (DetNum NumSg) (PossNP (UseN aunt_paternal_N) (DetCN (DetNum NumSg) (UseN man_N)))) (ComplV2 want_V2 (DetCN (DetNum NumSg) (UseN cheese_N)))))) NoVoc')
        ],
        "children": [
            [pgf.readExpr('NoPConj'),pgf.readExpr('UttS (UseCl TPresTemp PPos (PredVP (DetCN (DetNum NumSg) (UseN boy_N)) (UseV walk_V)))'),pgf.readExpr('NoVoc')],
            [pgf.readExpr('NoPConj'),pgf.readExpr('UttS (UseCl TFutTemp PPos (PredVP (DetCN (DetNum NumSg) (PossNP (UseN aunt_paternal_N) (DetCN (DetNum NumSg) (UseN man_N)))) (ComplV2 want_V2 (DetCN (DetNum NumSg) (UseN cheese_N)))))'),pgf.readExpr('NoVoc')]
        ],
        "cn_subtrees": [
            [pgf.readExpr('UseN boy_N')],
            [pgf.readExpr('PossNP (UseN aunt_paternal_N) (DetCN (DetNum NumSg) (UseN man_N))'),pgf.readExpr('UseN aunt_paternal_N'),pgf.readExpr('UseN man_N'),pgf.readExpr('UseN cheese_N')]
        ],
        "cn_subtrees_nonoverlap": [
            [pgf.readExpr('UseN boy_N')],
            [pgf.readExpr('PossNP (UseN aunt_paternal_N) (DetCN (DetNum NumSg) (UseN man_N))'),pgf.readExpr('UseN cheese_N')]
        ],
        "usen_subtrees": [
            [pgf.readExpr('UseN boy_N')],
            [pgf.readExpr('UseN aunt_paternal_N'),pgf.readExpr('UseN man_N'),pgf.readExpr('UseN cheese_N')]
        ],
        "leaf_cats": {
            "NumSg": ['NumSg', 'NumPl'],
            "TPresTemp": ['TPresTemp', 'TFutTemp', 'TPastTemp', 'TRemFutTemp', 'TRemPastTemp']
        },
        "leaf_nodes_with_ids": [
            [(pgf.readExpr("NoPConj"), 0), (pgf.readExpr("TPresTemp"), 1), (pgf.readExpr("PPos"), 2), (pgf.readExpr("NumSg"), 3), (pgf.readExpr("boy_N"), 5), (pgf.readExpr("walk_V"), 8), (pgf.readExpr("NoVoc"), 13)],
            [(pgf.readExpr("NoPConj"), 0),
              (pgf.readExpr("TFutTemp"), 1),
              (pgf.readExpr("PPos"), 2),
              (pgf.readExpr("NumSg"), 3),
              (pgf.readExpr("aunt_paternal_N"), 5),
              (pgf.readExpr("NumSg"), 7),
              (pgf.readExpr("man_N"), 9),
              (pgf.readExpr("want_V2"), 14),
              (pgf.readExpr("NumSg"), 15),
              (pgf.readExpr("cheese_N"), 17),
              (pgf.readExpr("NoVoc"), 24)]
        ],
        "exprs_with_empty_nodes": [
            (pgf.readExpr('PhrUtt NoPConj (UttS (UseCl TPresTemp ? (PredVP (DetCN (DetNum NumSg) (UseN boy_N)) ?))) NoVoc'),(2,8)),
            (pgf.readExpr('PhrUtt NoPConj (UttS (UseCl TFutTemp PPos (PredVP (DetCN (DetNum NumSg) (PossNP (UseN aunt_paternal_N) ?)) (ComplV2 want_V2 (DetCN (DetNum NumSg) (UseN cheese_N)))))) ?'),(7,29))
        ],
        "substitute_nodes": [
            [pgf.readExpr("PNeg"),pgf.readExpr("(CopNPAssoc (PronPostdetNP it15_Pron (DemPostdet that_Quant)))")],
            [pgf.readExpr("(PostdetCN (PredetN (QuantPredet all_QuantStem) hand_N) (QuantDemPostdet all_QuantStem that_Quant) (DetNum NumSg))"),pgf.readExpr("NoVoc")]
        ],
        "reconstituted_tree_strs": [
            "PhrUtt NoPConj (UttS (UseCl TPresTemp PNeg (PredVP (DetCN (DetNum NumSg) (UseN boy_N)) (CopNPAssoc (PronPostdetNP it15_Pron (DemPostdet that_Quant)))))) NoVoc",
            "PhrUtt NoPConj (UttS (UseCl TFutTemp PPos (PredVP (DetCN (DetNum NumSg) (PossNP (UseN aunt_paternal_N) (PostdetCN (PredetN (QuantPredet all_QuantStem) hand_N) (QuantDemPostdet all_QuantStem that_Quant) (DetNum NumSg)))) (ComplV2 want_V2 (DetCN (DetNum NumSg) (UseN cheese_N)))))) NoVoc"
        ],
        "empty_leaves_tree_strs": [
            "PhrUtt ? (UttS (UseCl ? ? (PredVP (DetCN (DetNum ?) (UseN ?)) (CopNPAssoc (PronPostdetNP ? (DemPostdet ?)))))) ?",
            "PhrUtt ? (UttS (UseCl ? ? (PredVP (DetCN (DetNum ?) (PossNP (UseN ?) (PostdetCN (PredetN (QuantPredet ?) ?) (QuantDemPostdet ? ?) (DetNum ?)))) (ComplV2 ? (DetCN (DetNum ?) (UseN ?)))))) ?"
        ],
        "children_types": {
            "UseN": ["N"],
            "PhrUtt": ["PConj","Utt","Voc"],
            "TPresTemp": []
        },
        "random_cats": [
            'N', 'Temp','Utt','V','VP'
        ],
        "patterns": [
            ('Phr',"PhrUtt NoPConj (UttS (UseCl TPresTemp ? (PredVP (DetCN (DetNum NumSg) (UseN boy_N)) (CopNPAssoc (PronPostdetNP it15_Pron ?))))) NoVoc"),
            ('Phr',"PhrUtt NoPConj (UttS (UseCl TFutTemp PPos (PredVP (DetCN (DetNum NumSg) (PossNP (UseN aunt_paternal_N) (PostdetCN (PredetN (QuantPredet all_QuantStem) hand_N) ? (DetNum NumSg)))) (ComplV2 want_V2 ?)))) NoVoc"),
            ('CN','PossNP (UseN aunt_paternal_N) (DetCN (DetNum NumSg) (UseN man_N))')
        ],
        "tokens": [
            ['umfana', 'u', '&+', 'ya', '&+', 'hamb', '&+', 'a'],
            ['ubabekazi', 'we', '&+', 'ndoda', 'u', '&+', 'zo', '&+', 'fun', '&+', 'a', 'ushizi']
        ],
        "segments": [
            ['umfana', '_', 'u', 'ya', 'hamb', 'a'],
            ['ubabekazi', '_', 'we', 'ndoda', '_', 'u', 'zo', 'fun', 'a', '_', 'ushizi']
        ],
        "leaves_tagged": [
            [('umfana', 'boy_N', 5),
                ('u', None, -1),
                ('&+', None, -1),
                ('ya', None, -1),
                ('&+', None, -1),
                ('hamb&+a', 'walk_V', 8)
            ],
            [('ubabekazi', 'aunt_paternal_N', 5),
                ('we', None, -1),
                ('&+', None, -1),
                ('ndoda', 'man_N', 9),
                ('u', None, -1),
                ('&+', None, -1),
                ('zo', None, -1),
                ('&+', None, -1),
                ('fun&+a', 'want_V2', 14),
                ('ushizi', 'cheese_N', 17)]
        ],
        "lin_brackets": [
            [
                (10,'umfana u&+ya&+hamb&+a'),
                (8,'hamb&+a'),
                (7,'umfana'),
                (12,'umfana u&+ya&+hamb&+a')
            ],
            [
                (12,'ubabekazi we&+ndoda'),
                (20,'u&+zo&+fun&+a ushizi'),
                (5,'ubabekazi'),
                (9,'ndoda'),
                (14,'fun&+a'),
                (17,'ushizi')
            ]
        ]
    }



# -*- coding: utf-8 -*-

import pytest
import pgf
import pgfaux.analyze as an

def test_node_is_empty(multilexlang_2022_07):
    data = multilexlang_2022_07
    for e in data['exprs']:
        assert an.node_is_empty(e) == False
    expr = pgf.readExpr('?')
    assert an.node_is_empty(expr) == True

def test_is_equal(multilexlang_2022_07):
    data = multilexlang_2022_07
    assert an.is_equal(data['exprs'][0],data['exprs'][0]) == True
    assert an.is_equal(data['exprs'][0],data['exprs'][1]) == False

def test_depth(multilexlang_2022_07):
    data = multilexlang_2022_07
    assert an.depth(data['exprs'][0]) == 6
    assert an.depth(data['exprs'][1]) == 8
    assert an.depth(pgf.readExpr('?')) == 0
    assert an.depth(pgf.readExpr('RootFun')) == 0

def test_root_str(multilexlang_2022_07):
    data = multilexlang_2022_07
    assert an.root_str(data['exprs'][0]) == 'PhrUtt'
    assert an.root_str(pgf.readExpr('?')) == None
    assert an.root_str(pgf.readExpr('"symbol"')) == '"symbol"'

def test_root_cat(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    for e in data['exprs']:
        assert an.root_cat(e,grammar) == 'Phr'
    assert an.root_cat(pgf.readExpr('"symbol"'),grammar) == an.SYMB

def test_children_trees(multilexlang_2022_07):
    data = multilexlang_2022_07
    for i in range(len(data['exprs'])):
        e = data['exprs'][i]
        cs = tuple([str(c) for c in data['children'][i]])
        cs_ = tuple([str(c) for c in an.children_trees(e)])
        assert cs == cs_
    assert an.children_trees(pgf.readExpr('?')) == []
    assert an.children_trees(pgf.readExpr('"symbol"')) == []
    assert an.children_trees(pgf.readExpr('RootFun')) == []

def test_subtrees_of_cat(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    for i in range(len(data['exprs'])):
        e = data['exprs'][i]
        ss = tuple([str(s) for s in data['cn_subtrees'][i]])
        ss_ = tuple([str(s) for s in an.subtrees_of_cat(e,'CN',grammar,overlap=True)])
        assert ss == ss_
    for i in range(len(data['exprs'])):
        e = data['exprs'][i]
        ss = tuple([str(s) for s in data['cn_subtrees_nonoverlap'][i]])
        ss_ = tuple([str(s) for s in an.subtrees_of_cat(e,'CN',grammar,overlap=False)])
        assert ss == ss_

def test_subtrees_of_fun(multilexlang_2022_07):
    data = multilexlang_2022_07
    for i in range(len(data['exprs'])):
        e = data['exprs'][i]
        ss = tuple([str(s) for s in data['usen_subtrees'][i]])
        ss_ = tuple([str(s) for s in an.subtrees_of_fun(e,'UseN')])
        assert ss == ss_

def test_alternative_leaf_function_names(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    for l in data["leaf_cats"].keys():
        e = pgf.readExpr(l)
        ls = an.alternative_leaf_function_names(e,grammar,data["conc_name"])
        assert ls == data["leaf_cats"][l]

def test_leaf_nodes_with_ids(multilexlang_2022_07):
    data = multilexlang_2022_07
    for i in range(len(data["exprs"])):
        e = data["exprs"][i]
        ls = an.leaf_nodes_with_ids(e)
        assert ls == data["leaf_nodes_with_ids"][i]

def empty_nodes_with_ids(multilexlang_2022_07):
    data = multilexlang_2022_07
    for e,ids in data["exprs"]:
        assert tuple(an.empty_node_ids(e)) == ids



# -*- coding: utf-8 -*-

import pytest
import pgf
import pgfaux.generate as gen

def test_replace_empty_nodes(multilexlang_2022_07):
    data = multilexlang_2022_07
    for i in range(len(data["exprs_with_empty_nodes"])):
        tree = data["exprs_with_empty_nodes"][i][0]
        nodes = data["substitute_nodes"][i]
        assert str(gen.replace_empty_nodes(tree,nodes)[0]) == data["reconstituted_tree_strs"][i]

def test_make_leaves_empty(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data["reconstituted_tree_strs"])):
        full_tree = data["reconstituted_tree_strs"][i]
        empty_leaves_tree = data["empty_leaves_tree_strs"][i]
        emptied_tree = gen.make_leaves_empty(full_tree,grammar,concrete)
        assert str(emptied_tree) == empty_leaves_tree

def test_children_types(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    for fun in data['children_types'].keys():
        assert tuple(gen.children_types(grammar,fun)) == tuple(data['children_types'][fun])

def test_generate_random_tree_by_cat(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    for cat in multilexlang_2022_07['random_cats']:
        tree = gen.generate_random_tree_by_cat(grammar,cat)
        assert grammar.checkExpr(tree,pgf.readType(cat))

def test_generate_random_tree_by_pattern(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    for cat,pattern in multilexlang_2022_07['patterns']:
        tree = gen.generate_random_tree_by_pattern(grammar,pattern)
        assert grammar.checkExpr(tree,pgf.readType(cat))

def test_replace_at_id(multilexlang_2022_07):
    data = multilexlang_2022_07
    for i in range(len(data["exprs_with_empty_nodes"])):
        expr,ids = data["exprs_with_empty_nodes"][i]
        sub_nodes = data["substitute_nodes"][i]
        for j in range(len(ids)):
            expr = gen.replace_at_id(expr,sub_nodes[j],ids[j])
        assert str(expr) == data["reconstituted_tree_strs"][i]



# -*- coding: utf-8 -*-

import pgf
import pgfaux.analyze as analyze
from pgfaux.exceptions import *

import random
import itertools

MAX_GR_ATTEMPTS = 100
MAX_GR_DEPTH = 10

def replace_empty_nodes(tree,nodes):
    tree = pgf.readExpr(str(tree))
    if analyze.depth(tree) == 0 and str(tree) == '?':
        if len(nodes) > 1:
            return pgf.readExpr(str(nodes[0])),nodes[1:]
        elif len(nodes) > 0:
            return pgf.readExpr(str(nodes[0])),[]
        else:
            return tree,[]
    children = analyze.children_trees(tree)
    new_children = []
    for c in children:
        new_c,nodes = replace_empty_nodes(c,nodes)
        new_children.append(new_c)
    return pgf.Expr(analyze.root_str(tree),new_children),nodes

def make_leaves_empty(tree,grammar,concrete):
    tree = pgf.readExpr(str(tree))
    depth = analyze.depth(tree)
    root_fun_name = analyze.root_str(tree)
    if depth == 0:
        if str(tree) == '?':
            return tree
        else:
            return pgf.readExpr('?')
    children = analyze.children_trees(tree)
    new_children = []
    for c in children:
        new_c = make_leaves_empty(c,grammar,concrete)
        new_children.append(new_c)
    return pgf.Expr(root_fun_name,new_children)

def children_types(grammar,fun_str):
    funtype = grammar.functionType(fun_str)
    if len(funtype.hypos) == 0:
        return []
    return [str(funtype.hypos[i][2]) for i in range(len(funtype.hypos))]

def generate_random_fun(grammar,cat_str):
    funs = grammar.functionsByCat(cat_str)
    print(cat_str,funs)
    index = random.randint(0,max(len(funs)-1,0))
    if index >= len(funs):
        return None
    return funs[index]

def _generate_random_tree_by_cat(grammar,cat_str,depth):
    fun_str = generate_random_fun(grammar,cat_str)
    if not fun_str:
        return None
    children = children_types(grammar,fun_str)
    if len(children) == 0:
        return pgf.readExpr(fun_str)
    elif depth == 0:
        return None
    else:
        children_trees = []
        for c in children:
            ct = _generate_random_tree_by_cat(grammar,c,depth-1)
            if not ct:
                return None
            children_trees.append(ct)
        return pgf.Expr(fun_str,children_trees)

def generate_random_tree_by_cat(grammar,cat_str,depth=MAX_GR_DEPTH):
    for i in range(MAX_GR_ATTEMPTS):
        tree = _generate_random_tree_by_cat(grammar,cat_str,depth)
        if tree:
            return tree

def generate_random_tree_by_pattern(grammar,pattern,depth=MAX_GR_DEPTH):
    pattern = pgf.readExpr(str(pattern))
    if analyze.node_is_empty(pattern):
        return generate_random_tree_by_cat(grammar,str(grammar.startCat),depth)
    else:
        fun,children = pattern.unpack()
        c_types = children_types(grammar,fun)
        if len(c_types) != len(children):
            raise InvalidPattern(pattern)
        new_children = []
        if len(children) > 0:
            for i in range(len(children)):
                if analyze.node_is_empty(children[i]):
                    new_children.append(generate_random_tree_by_cat(grammar,c_types[i],depth-1))
                else:
                    new_children.append(generate_random_tree_by_pattern(grammar,children[i],depth-1))
            return pgf.Expr(fun,new_children)
        else:
            return pgf.readExpr(fun)

def _replace_at_id(tree,subtree,id,prev_id=-1):
    tree = pgf.readExpr(str(tree))
    children = analyze.children_trees(tree)
    if len(children) == 0: # tree is a leaf
        cur_id = prev_id + 1
        if cur_id == id:
            return subtree,cur_id
        else:
            return tree,cur_id
    new_children = []
    for c in children:
        new_c,prev_id = _replace_at_id(c,subtree,id,prev_id)
        new_children.append(new_c)
        cur_id = prev_id + 1
    if cur_id == id:
        return subtree,cur_id
    return pgf.Expr(analyze.root_str(tree),new_children),cur_id

def replace_at_id(tree,subtree,id):
    new_tree,root_id = _replace_at_id(tree,subtree,id)
    return new_tree



# -*- coding: utf-8 -*-

import pytest
import pgf
import pgfaux.linearize as lin

def test_tokens_from_bracket(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        [b] = concrete.bracketedLinearize(data['exprs'][i])
        assert tuple(lin.tokens_from_bracket(b)) == tuple(data['tokens'][i])

def test_segments_from_bracket(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        [b] = concrete.bracketedLinearize(data['exprs'][i])
        assert tuple(lin.segments_from_bracket(b)) == tuple(data['segments'][i])

def test_tokens_from_expr(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        assert tuple(lin.tokens_from_expr(data['exprs'][i],concrete)) == tuple(data['tokens'][i])

def test_segments_from_expr(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        assert tuple(lin.segments_from_expr(data['exprs'][i],concrete)) == tuple(data['segments'][i])

def test_leaf_tagged_linearize_bracket(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        [b] = concrete.bracketedLinearize(data['exprs'][i])
        assert tuple(lin.leaf_tagged_linearize_bracket(b)) == tuple(data['leaves_tagged'][i])

def test_leaf_tagged_linearize(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        assert tuple(lin.leaf_tagged_linearize(data['exprs'][i],concrete)) == tuple(data['leaves_tagged'][i])

def test_lin_at_fid_bracket(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        e = data['exprs'][i]
        [b] = concrete.bracketedLinearize(e)
        for (fid,l) in data['lin_brackets'][i]:
            assert lin.lin_at_fid_bracket(b,fid) == l

def test_lin_at_fid(multilexlang_2022_07):
    data = multilexlang_2022_07
    grammar = pgf.readPGF(data['grammar'])
    concrete = grammar.languages[data['conc_name']]
    for i in range(len(data['exprs'])):
        e = data['exprs'][i]
        for (fid,l) in data['lin_brackets'][i]:
            assert lin.lin_at_fid(e,concrete,fid) == l



# -*- coding: utf-8 -*-

import pgf

BIND = '&+'
WORD_BOUNDARY = '_'

def concat_bind(bracket_elements,bind=BIND):
    c_str = ""
    for e in bracket_elements:
        if isinstance(e,type(pgf.BIND)):
            c_str += bind
        else:
            c_str += str(e)
    return c_str

def is_leaf_bracket(bracket):
    for c in bracket.children:
        if isinstance(c,pgf.Bracket):
            return False
    return True

def tokens_from_bracket(bracket,bind=BIND):
    toks = []
    for b in bracket.children:
        if isinstance(b,str):
            toks.append(b)
        elif isinstance(b,type(pgf.BIND)):
            toks.append(bind)
        else:
            c_toks = tokens_from_bracket(b,bind)
            toks += c_toks
    return toks

def segments_from_bracket(bracket,wb=WORD_BOUNDARY):
    toks = tokens_from_bracket(bracket,BIND)
    segs = []
    for i in range(len(toks)-1):
        if toks[i] == BIND:
            continue
        elif toks[i+1] == BIND:
            segs.append(toks[i])
        else:
            segs += [toks[i],wb]
    segs.append(toks[-1])
    return segs

def tokens_from_expr(expr,conc,bind=BIND):
    bl = conc.bracketedLinearize(expr)
    root_bracket = bl[0]
    return tokens_from_bracket(root_bracket,bind)

def segments_from_expr(expr,conc,wb=WORD_BOUNDARY):
    bl = conc.bracketedLinearize(expr)
    root_bracket = bl[0]
    return segments_from_bracket(root_bracket,wb)

def leaf_tagged_linearize_bracket(bracket,bind=BIND):
    lin_pieces = []
    if isinstance(bracket,str):
        return [(bracket,None,-1)]
    if isinstance(bracket,type(pgf.BIND)):
    # if str(bracket) == BIND:
        bracket = (bind,None,-1)
        return [bracket]
    if is_leaf_bracket(bracket):
        lin_pieces.append((concat_bind(bracket.children,bind),bracket.fun,bracket.fid))
    else:
        for c in bracket.children:
            c_pieces = leaf_tagged_linearize_bracket(c,bind)
            lin_pieces += c_pieces
    return lin_pieces

def leaf_tagged_linearize(tree,concrete,bind=BIND):
    tree = pgf.readExpr(str(tree))
    [root_bracket] = concrete.bracketedLinearize(tree)
    return leaf_tagged_linearize_bracket(root_bracket,bind)

def concat_tokens(tokens,bind=BIND):
    pairs = [(None,tokens[0])] + [(tokens[i-1],tokens[i]) for i in range(1,len(tokens))]
    lin = ''
    for (prev,cur) in pairs:
        if prev and not prev == bind and not cur == bind:
            lin += ' '
        lin += cur
    return lin

def lin_at_fid_bracket(bracket,id,bind=BIND):
    if not isinstance(bracket,pgf.Bracket):
        return None
    bracket_children = [b for b in bracket.children if isinstance(b,pgf.Bracket)]
    if len(bracket_children) == 0 and bracket.fid != id:
        return None

    if bracket.fid == id:
        b_lins = []
        for b in bracket.children:
            if isinstance(b,str):
                b_lins.append(b)
            elif isinstance(b,type(pgf.BIND)):
                b_lins.append(bind)
            elif isinstance(b,pgf.Bracket):
                b_lins.append(lin_at_fid_bracket(b,b.fid,bind))
        return concat_tokens(b_lins,bind)
    else:
        for b in bracket.children:
            b_lin = lin_at_fid_bracket(b,id,bind)
            if b_lin:
                return b_lin
    return None

def lin_at_fid(tree,concrete,id,bind=BIND,prev_id=-1):
    tree = pgf.readExpr(str(tree))
    [b] = concrete.bracketedLinearize(tree)
    return lin_at_fid_bracket(b,id,bind)
